/***************************************************************************
               ulxr_xmlparse.h  -  parse xml files
                             -------------------
    begin                : Wed Mar 13 2002
    copyright            : (C) 2002-2004 by Ewald Arnold
    email                : ulxmlrpcpp@ewald-arnold.de

    $Id: ulxr_xmlparse.h,v 1.8 2004/01/02 17:47:30 ewald-arnold Exp $


 ***************************************************************************/

/**************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ***************************************************************************/

#ifndef ULXR_XMLPARSE
#define ULXR_XMLPARSE

#include <ulxmlrpcpp.h>  // always first header
#include <ulxr_expatwrap.h>

#include <stack>


namespace ulxr {


/** Base class for XML parsing.
  */
class ULXR_EXPORT XmlParser : public ExpatWrapper
{
 protected:

 /** Destroys the parser.
   * The derived class is responsible for cleaning up stack<ParserState*>.
   */
   virtual ~XmlParser() = 0;

 public:

 /** Constructs a parser.
   */
   XmlParser();

 /** Tests if parsing has completed.
   * Completed means that all opening tags have been correctly closed.
   * @return true if completed
   */
   bool isComplete() const;

 /** Sets the complete state.
   * @param  true if file read
   */
   void setComplete(bool comp);

   enum State { eNone,               //!<  state after start
                eUnknown,            //!<  unknwon state when an error occured
                eXmlParserLast       //!<  used to chain next parser class
              };

/** Helper class to represent the data of the current parsing step.
  */
  class ULXR_EXPORT ParserState
  {
   public:

   /** Constructs a ParserState.
     * @param  the actual ParserState
     */
     ParserState (unsigned st);

   /** Destroys the ParserState.
     */
     virtual ~ParserState();

   /** Gets the ParserState of this ParserState
     * @return the actual ParserState
     */
     unsigned getParserState() const;

   /** Gets the privious ParserState
     * @return the previous ParserState
     */
     unsigned getPrevParserState() const;

   /** Sets the privious ParserState
     * @param  prev the previous ParserState
     */
     void setPrevParserState(unsigned prev);

   /** Gets the name of the ParserState.
     * Useful only for debugging.
     * @return the name of actual ParserState
     */
     virtual CppString getStateName() const;

   /** Appends some characters of the ParserState.
     * This is a part of the data of an xml rpc element.
     * @param  s   the current chunk of text
     * @param  len valid len.
     */
     void appendCharData(const XML_Char *s, int len);

   /** Gets the characters of the ParserState.
     * @return  the data element
     */
     CppString getCharData() const;

   private:

     ParserState(const ParserState&); // forbid this
     ParserState& operator= (const ParserState&);

     unsigned   state;
     unsigned   prevstate;
     CppString  cdata;
  };

 protected:

 /** Checks if current and expected tag are the same.
   * @param  current   the name of the current tag
   * @param  expected  name of the expected tag
   */
   void assertEndElement(const char *current, const char *expected);

 /** Tests if the current opening tag is to be parsed by this
   * inheritance level or by the parent.
   * Used ONLY internally.
   * @param  name  the name of the current tag
   * @param  atts  pointer to the current attributes (unused in XML-RPC)
   */
   bool testStartElement(const XML_Char *name, const XML_Char **atts);

 /** Tests if the current closing tag is to be parsed by this
   * inheritance level or by the parent.
   * Used ONLY internally.
   * @param  name  the name of the current tag
   * @param  pointer to the current attributes (unused in XML-RPC)
   */
   bool testEndElement(const XML_Char *name);

   std::stack<ParserState*>  states;

private:

 /** Parses the content of the current xml element.
   * Used ONLY internally as callback from expat.
   * The text from expat is encoded in UTF8.
   * @param  s   the current chunk of text
   * @param  len valid len.
   */
   virtual void charData(const XML_Char *s, int len);

   bool complete;
};


};  // namespace ulxr


#endif // ULXR_XMLPARSE
